/*----------------------------------------------------------------
	author: massimo.sala.71@gmail.com
	license: MIT
	pls report bugs/improvements to me
*/

#include "config.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#define safe_strlen(sz)		(sz == NULL ? 0 : strlen(sz))

#include <stdbool.h>
#include <stdlib.h>						// secure_getenv

#include "chk_string_mask.h"
#include "protocol.h"
#include "md5.h"
#include "exceptions/AssertException.h"
#include "exceptions/IOException.h"
#include "exceptions/MemoryException.h"
#include "exceptions/ProtocolException.h"

/*
https://www.postgresql.org/docs/8.4/protocol-message-formats.html
https://www.postgresql.org/docs/8.4/protocol-error-fields.html
https://www.postgresql.org/docs/12/protocol-changes.html
*/
#define PROTO_V3			3

#define PKT_Error			'E'
#define PKT_AuthRequest		'R'
#define PKT_Password		'p'
#define PKT_Terminate		'X'

#define AUTH_Plain			3
#define AUTH_MD5			5


const char * safe_getenv(const char * szName, const char * szDefault)
{
	const char * tmp;

	if (szName == NULL)
		return	NULL;

#ifdef secure_getenv						// glibc 2.17
	tmp = secure_getenv(szName);
#else
	tmp = getenv(szName);
#endif
	if (tmp == NULL || *tmp == 0)
		return	szDefault;
	else
		return	tmp;	
}


void pgsql_md5(char * output, const char * data1, int len1, const char * data2, int len2)
{
	static char hex[] = "0123456789abcdef";
	md5_context_t ctx;
	md5_byte_t hash[16];

	md5_init(&ctx);
	md5_append(&ctx, (const md5_byte_t *) data1, len1);
	md5_append(&ctx, (const md5_byte_t *) data2, len2);
	md5_finish(&ctx, hash);
	for (int i = 0; i < 16; i++)
	{
		*output ++ = hex[hash[i] >> 4];
		*output ++ = hex[hash[i] & 0xf];
	}
}


/**
 *  PostgreSQL test. Default port: 5432/tcp
 *
 *  Optional arguments: database username password
 *  If the arguments aren't declared, read them from the environment variables:
 *    PGDATABASE, PGUSER, PGPASSWORD
 *  Default values: database = postgres, username = postgres
 *
 *  @file
 */
void check_pgsql(Socket_T socket)
{
	Port_T port;
	const char * szDatabase	= NULL;
	const char * szUsername	= NULL;
	const char * szPassword	= NULL;

	#define LOGIN_MAXLEN		255			//!!! beware this limits the length of the credentials
	char requestLogin[LOGIN_MAXLEN];

	#define REPLY_MINLEN		9
	#define REPLY_MAXLEN		1024
    char reply[REPLY_MAXLEN + 1];
	int br;

	#define PKT_AuthRequestLen	9
	char responseAuthOk[PKT_AuthRequestLen] =
	{
		PKT_AuthRequest,
		0x00,                              /** Length */
		0x00,
		0x00,
		0x08,
		0x00,                              /** OK code 0 */
		0x00,
		0x00,
		0x00
	};

	#define PKT_requestTerminateLen	5
	char requestTerminate[PKT_requestTerminateLen] =
	{
		PKT_Terminate,
		0x00,                              /** Length */
		0x00,
		0x00,
		0x04
	};

	size_t iLoginLen;
	char * ptr;
	bool bAuthConnected = false;
	bool bReplyHasError = false;

	ASSERT(socket);

	port = Socket_getPort(socket);
	if (port != NULL)
	{
		szDatabase = port->parameters.pgsql.database;
		szUsername = port->parameters.pgsql.username;
		szPassword = port->parameters.pgsql.password;
	}
	if (szDatabase == NULL || * szDatabase == 0)
	{
		DEBUG("PGSQL: datatase not specified, reading environment PGDATABASE\n");
		szDatabase = safe_getenv("PGDATABASE", "postgres");
	}
	if (szUsername == NULL || * szUsername == 0)
	{
		DEBUG("PGSQL: username not specified, reading environment PGUSER\n");
		szUsername = safe_getenv("PGUSER", "postgres");
	}
	if (szPassword == NULL)
	{
		DEBUG("PGSQL: password not specified, reading environment PGPASSWORD\n");
		szPassword = safe_getenv("PGPASSWORD", "");
	}

	iLoginLen = 8 + 4 + 1 + safe_strlen(szUsername) + 1 + 8 + 1 + safe_strlen(szDatabase) + 1 + 1;
	if (iLoginLen > LOGIN_MAXLEN)
		THROW(MemoryException, "PGSQL: internal error (insufficient buffer for database + username) -- %s", STRERROR);
// server expects authentication: AUTH_Plain
	if (4 + safe_strlen(szPassword) > LOGIN_MAXLEN)
		THROW(MemoryException, "PGSQL: internal error (insufficient buffer for password) -- %s", STRERROR);

	memset(requestLogin, 0, LOGIN_MAXLEN);					// set last NULL after payload
	ptr = requestLogin + 8;
	ptr = strpush_szstr_null(ptr, "user");
	ptr = strpush_szstr_null(ptr, szUsername);
	ptr = strpush_szstr_null(ptr, "database");
	ptr = strpush_szstr_null(ptr, szDatabase);
	++ ptr;													// trailing NULL

	if (iLoginLen != ptr - requestLogin)
		THROW(AssertException, "PGSQL: internal error (request packet) -- %s", STRERROR);

	requestLogin[3] = (unsigned char) iLoginLen;
	requestLogin[5] = PROTO_V3;

	Socket_nodelay(socket);
	if (Socket_write(socket, requestLogin, iLoginLen) != iLoginLen)
		THROW(IOException, "PGSQL: error sending PGSQL login request -- %s", STRERROR);

	br = Socket_read(socket, reply, REPLY_MAXLEN);
	if (br < REPLY_MINLEN)
		THROW(IOException, "PGSQL: error receiving data -- %s", STRERROR);

	if (*reply == PKT_Error)
		THROW(ProtocolException, "PGSQL: server replies Error -- %s", STRERROR);

// server expects authentication
	if (br >= PKT_AuthRequestLen && memcmp(reply, responseAuthOk, PKT_AuthRequestLen) != 0)
	{
		memset(requestLogin, 0, LOGIN_MAXLEN);			// set last NULL after payload
		requestLogin[0] = PKT_Password;
		if (br == PKT_AuthRequestLen && reply[4] == br-1 && reply[8] == AUTH_Plain)
		{
			DEBUG("PGSQL: server expect password, AUTH_Plain\n");
			iLoginLen = 4 + safe_strlen(szPassword);
			requestLogin[4] = (unsigned char) iLoginLen;
			strcpy(requestLogin + 5, szPassword);
			if (Socket_write(socket, requestLogin, iLoginLen + 1) <= 0)
				THROW(IOException, "PGSQL: error sending AUTH Plain response -- %s", STRERROR);
		}
		else
			if (br == PKT_AuthRequestLen + 4 && reply[4] == br-1 && reply[8] == AUTH_MD5)
			{
				DEBUG("PGSQL: server expect password, AUTH_MD5\n");
				iLoginLen = 8 + 32;
				requestLogin[4] = (unsigned char) iLoginLen;
				strcpy(requestLogin + 5, "md5");
//!!! first char is PKT_Password
//				char pkt_hash[8 + 32 + 1] = "p\0\0\0\x28md5                                \0";
				pgsql_md5(requestLogin + 8, szPassword, safe_strlen(szPassword), szUsername, safe_strlen(szUsername));
				pgsql_md5(requestLogin + 8, requestLogin + 8, 32, reply + 9, 4);
				if (Socket_write(socket, requestLogin, iLoginLen + 1) <= 0)
					THROW(IOException, "PGSQL: error sending AUTH MD5 response -- %s", STRERROR);
			}
			else
				THROW(ProtocolException, "PGSQL: unknown authentication request -- %s", STRERROR);

		br = Socket_read(socket, reply, REPLY_MAXLEN);
		if (br < REPLY_MINLEN)
			THROW(IOException, "PGSQL: error receiving data -- %s", STRERROR);

//!!! PKT_Error: FATAL C28000 password authentication failed for user...
	}


// successful connection? with or without authentication?
	if (br < PKT_AuthRequestLen || memcmp(reply, responseAuthOk, PKT_AuthRequestLen) != 0)
		ptr = reply;
	else
	{
		bAuthConnected = true;
		ptr = &reply[PKT_AuthRequestLen];
		br -= PKT_AuthRequestLen;
	}

	ptr[br] = 0;
	if (* ptr == PKT_Error)
	{
		bReplyHasError = true;
		ptr += 1 + 4;
		br -= 1 + 4;
	}

	for (int i = 0; i < br; i ++)
	{
		char ch = ptr[i];
		if (ch < ' ' || ch > 0x7e)
			ptr[i] = ' ';				// remove unprintable chars
		if (ch == 0 && i + 1 < br)
			ptr[i + 1] = ' ';			// remove Type before the text message
	}

/*
PROTST --> PROTST
PGSQL: folium dbread  CONN= 0 ERR= 1 : SFATAL C28000 Mpassword authentication failed for user "dbread" Fauth.c L273 Rauth_failed   -- Resource temporarily unavailable

PROTST --> localhost
PGSQL: postgres2 postgres  CONN= 1 ERR= 1 : SFATAL C3D000 Mdatabase "postgres2" does not exist Fpostinit.c L422 RInitPostgres   -- Resource temporarily unavailable
*/

// polite: close connection (harmless if server already closed)
	Socket_write(socket, requestTerminate, PKT_requestTerminateLen);

	if (bReplyHasError || ! bAuthConnected)
		THROW(ProtocolException, "PGSQL: %s %s : %s -- %s", szDatabase, szUsername, ptr, STRERROR);
}
